#  You may distribute under the terms of either the GNU General Public License
#  or the Artistic License (the same terms as Perl itself)
#
#  (C) Paul Evans, 2021 -- leonerd@leonerd.org.uk

package Future::Workflow 0.01;

use v5.14;
use warnings;

=head1 NAME

C<Future::Workflow> - asynchronous workflow management using futures

=head1 DESCRIPTION

This distribution is a work-in-progress, aiming to eventually contain a
collection of classes for arranging L<Future>-based workflow patterns into
larger structures.

=cut

=head1 AUTHOR

Paul Evans <leonerd@leonerd.org.uk>

=cut

0x55AA;
